@@ -3,9 +3,6 @@
 
 #include "settings.h"
 
-#include "logging_categories_p.h"
-#include "util.h"
-
 #include <QtCore/QUrl>
 
 using namespace Quotient;
@@ -21,11 +18,7 @@ void Settings::setLegacyNames(const QString& organizationName,
 }
 
 Settings::Settings(QObject* parent) : QSettings(parent)
-{
-#if (QT_VERSION < QT_VERSION_CHECK(6, 0, 0))
-    setIniCodec("UTF-8");
-#endif
-}
+{}
 
 void Settings::setValue(const QString& key, const QVariant& value)
 {
@@ -49,7 +42,7 @@ QVariant Settings::value(const QString& key, const QVariant& defaultValue) const
     // (QVariant("false") == true in JavaScript). Since we have a mixed
     // environment where both QSettings and Qt.labs.Settings may potentially
     // work with same settings, better ensure compatibility.
-    return value.toString() == QStringLiteral("false") ? QVariant(false) : value;
+    return value.toString() == "false"_L1 ? QVariant(false) : value;
 }
 
 bool Settings::contains(const QString& key) const
@@ -69,18 +62,14 @@ QStringList Settings::childGroups() const
 
 void SettingsGroup::setValue(const QString& key, const QVariant& value)
 {
-    Settings::setValue(groupPath + u'/' + key, value);
+    Settings::setValue(fullPath(key), value);
 }
 
-bool SettingsGroup::contains(const QString& key) const
-{
-    return Settings::contains(groupPath + u'/' + key);
-}
+bool SettingsGroup::contains(const QString& key) const { return Settings::contains(fullPath(key)); }
 
-QVariant SettingsGroup::value(const QString& key,
-                              const QVariant& defaultValue) const
+QVariant SettingsGroup::value(const QString& key, const QVariant& defaultValue) const
 {
-    return Settings::value(groupPath + u'/' + key, defaultValue);
+    return Settings::value(fullPath(key), defaultValue);
 }
 
 QString SettingsGroup::group() const { return groupPath; }
@@ -111,9 +100,8 @@ QUO_DEFINE_SETTING(AccountSettings, bool, keepLoggedIn, "keep_logged_in", false,
                    setKeepLoggedIn)
 
 namespace {
-constexpr auto HomeserverKey = "homeserver"_ls;
-constexpr auto AccessTokenKey = "access_token"_ls;
-constexpr auto EncryptionAccountPickleKey = "encryption_account_pickle"_ls;
+constexpr auto HomeserverKey = "homeserver"_L1;
+constexpr auto EncryptionAccountPickleKey = "encryption_account_pickle"_L1;
 }
 
 QUrl AccountSettings::homeserver() const
@@ -128,23 +116,15 @@ void AccountSettings::setHomeserver(const QUrl& url)
 
 QString AccountSettings::userId() const { return group().section(u'/', -1); }
 
-void AccountSettings::clearAccessToken()
-{
-    legacySettings.remove(AccessTokenKey);
-    legacySettings.remove(QStringLiteral("device_id")); // Force the server to
-                                                        // re-issue it
-    remove(AccessTokenKey);
-}
-
 QByteArray AccountSettings::encryptionAccountPickle()
 {
-    return value("encryption_account_pickle"_ls, QString()).toByteArray();
+    return value("encryption_account_pickle"_L1, QString()).toByteArray();
 }
 
 void AccountSettings::setEncryptionAccountPickle(
     const QByteArray& encryptionAccountPickle)
 {
-    setValue("encryption_account_pickle"_ls, encryptionAccountPickle);
+    setValue("encryption_account_pickle"_L1, encryptionAccountPickle);
 }
 
 void AccountSettings::clearEncryptionAccountPickle()
