@@ -12,13 +12,18 @@ namespace Quotient {
 
 class Connection;
 
-std::shared_ptr<Connection> createTestConnection(const QString& localUserName,
-                                                 const QString& secret,
-                                                 const QString& deviceName);
-
+std::shared_ptr<Connection> createTestConnection(QLatin1StringView localUserName,
+                                                 QLatin1StringView secret,
+                                                 QLatin1StringView deviceName);
 }
 
 #define CREATE_CONNECTION(VAR, USERNAME, SECRET, DEVICE_NAME)             \
     const auto VAR = createTestConnection(USERNAME, SECRET, DEVICE_NAME); \
     if (!VAR)                                                             \
         QFAIL("Could not set up test connection");
+
+inline bool waitForFuture(const auto& ft)
+    requires requires { ft.isFinished(); }
+{
+    return QTest::qWaitFor([ft] { return ft.isFinished(); }, 40000);
+}
