@@ -26,7 +26,6 @@
 #include <QtConcurrent/QtConcurrent>
 #include <QtNetwork/QNetworkReply>
 
-#include <functional>
 #include <iostream>
 
 using namespace Quotient;
@@ -43,7 +42,7 @@ private:
     void onNewRoom(Room* r);
     void doTests();
     void conclude();
-    void finalize();
+    void finalize(const QString& lastWords = {});
 
 private:
     Connection* c = nullptr;
@@ -170,13 +169,13 @@ void TestSuite::finishTest(const TestToken& token, bool condition,
     if (condition) {
         clog << item << " successful" << endl;
         if (targetRoom)
-            targetRoom->postMessage(origin % ": "_ls % QString::fromUtf8(item) % " successful"_ls,
+            targetRoom->postMessage(origin % ": "_L1 % QString::fromUtf8(item) % " successful"_L1,
                                     MessageEventType::Notice);
     } else {
         clog << item << " FAILED at " << file << ":" << line << endl;
         if (targetRoom)
-            targetRoom->postPlainText(origin % ": "_ls % QString::fromUtf8(item) % " FAILED at "_ls
-                                      % QString::fromUtf8(file) % ", line "_ls % QString::number(line));
+            targetRoom->postPlainText(origin % ": "_L1 % QString::fromUtf8(item) % " FAILED at "_L1
+                                      % QString::fromUtf8(file) % ", line "_L1 % QString::number(line));
     }
 
     emit finishedItem(item, condition);
@@ -186,34 +185,60 @@ TestManager::TestManager(int& argc, char** argv)
     : QCoreApplication(argc, argv), c(new Connection(this))
 {
     Q_ASSERT(argc >= 5);
+    // NOLINTBEGIN(cppcoreguidelines-pro-bounds-pointer-arithmetic)
     clog << "Connecting to Matrix as " << argv[1] << endl;
-    c->loginWithPassword(QString::fromUtf8(argv[1]), QString::fromUtf8(argv[2]), QString::fromUtf8(argv[3]));
+    c->loginWithPassword(QString::fromUtf8(argv[1]), QString::fromUtf8(argv[2]),
+                         QString::fromUtf8(argv[3]));
     targetRoomName = QString::fromUtf8(argv[4]);
-    clog << "Test room name: " << argv[4] << endl;
+    clog << "Test room name: " << argv[4] << '\n';
     if (argc > 5) {
         origin = QString::fromUtf8(argv[5]);
-        clog << "Origin for the test message: " << origin.toStdString() << endl;
+        clog << "Origin for the test message: " << origin.toStdString() << '\n';
     }
+    clog.flush();
+    // NOLINTEND(cppcoreguidelines-pro-bounds-pointer-arithmetic)
 
-    connect(c, &Connection::connected, this, &TestManager::setupAndRun);
+    connect(c, &Connection::connected, this, [this] {
+        if (QUO_ALARM(c->homeserver().isEmpty() || !c->homeserver().isValid())
+            || QUO_ALARM(c->domain() != c->userId().section(u':', 1))) {
+            clog << "Connection information doesn't look right, "
+                 << "check the parameters passed to quotest" << endl;
+            QCoreApplication::exit(-2);
+            return;
+        }
+        clog << "Connected, server: " << c->homeserver().toDisplayString().toStdString() << '\n'
+             << "Access token: " << c->accessToken().toStdString() << endl;
+
+        // Test Connection::assumeIdentity() while we can replace connection objects
+        auto* newC = new Connection(c->homeserver(), this);
+        newC->assumeIdentity(c->userId(), c->deviceId(), QString::fromLatin1(c->accessToken()));
+        // NB: this will need to change when we switch E2EE on in quotest because encryption
+        //     data is initialised asynchronously
+        if (QUO_ALARM(newC->homeserver() != c->homeserver())
+            || QUO_ALARM(newC->userId() != c->userId()) || QUO_ALARM(!newC->isLoggedIn())) {
+            clog << "Connection::assumeIdentity() is broken" << endl;
+            QCoreApplication::exit(-2);
+            return;
+        }
+
+        c->deleteLater();
+        c = newC;
+        setupAndRun();
+    });
     connect(c, &Connection::resolveError, this,
         [](const QString& error) {
-            clog << "Failed to resolve the server: " << error.toStdString()
-                 << endl;
+            clog << "Could not start testing: " << error.toStdString() << endl;
             QCoreApplication::exit(-2);
         },
         Qt::QueuedConnection);
     connect(c, &Connection::loginError, this,
         [this](const QString& message, const QString& details) {
-            clog << "Failed to login to "
-                 << c->homeserver().toDisplayString().toStdString() << ": "
-                 << message.toStdString() << endl
-                 << "Details:" << endl
+            clog << "Failed to login to " << c->homeserver().toDisplayString().toStdString() << ": "
+                 << message.toStdString() << "\nDetails:\n"
                  << details.toStdString() << endl;
             QCoreApplication::exit(-2);
         },
         Qt::QueuedConnection);
-    connect(c, &Connection::loadedRoomState, this, &TestManager::onNewRoom);
 
     // Big countdown watchdog
     QTimer::singleShot(180000, this, [this] {
@@ -229,18 +254,23 @@ void TestManager::setupAndRun()
 {
     Q_ASSERT(!c->homeserver().isEmpty() && c->homeserver().isValid());
     Q_ASSERT(c->domain() == c->userId().section(u':', 1));
-    clog << "Connected, server: "
-         << c->homeserver().toDisplayString().toStdString() << endl;
-    clog << "Access token: " << c->accessToken().toStdString() << endl;
+    clog << "Connected, server: " << c->homeserver().toDisplayString().toStdString() << '\n'
+         << "Access token: " << c->accessToken().toStdString() << endl;
+
+    connect(c, &Connection::loadedRoomState, this, &TestManager::onNewRoom);
 
     c->setLazyLoading(true);
 
     clog << "Joining " << targetRoomName.toStdString() << endl;
-    auto joinJob = c->joinRoom(targetRoomName);
-    // Ensure that the room has been joined and filled with some events
-    // so that other tests could use that
-    connect(joinJob, &BaseJob::success, this, [this, joinJob] {
-        testSuite = new TestSuite(c->room(joinJob->roomId()), origin, this);
+    c->joinAndGetRoom(targetRoomName).then(this, [this](Room* room) {
+        if (!room) {
+            clog << "Failed to join the test room" << endl;
+            finalize();
+            return;
+        }
+        // Ensure that the room has been joined and filled with some events
+        // so that other tests could use that
+        testSuite = new TestSuite(room, origin, this);
         // Only start the sync after joining, to make sure the room just
         // joined is in it
         c->syncLoop();
@@ -255,32 +285,23 @@ void TestManager::setupAndRun()
             }
             if (!running.empty()) {
                 clog << running.size() << " test(s) in the air:";
-                for (const auto& test: qAsConst(running))
+                for (const auto& test: std::as_const(running))
                     clog << " " << testName(test);
                 clog << endl;
             }
             if (i == 1) {
-                testSuite->room()->getPreviousContent();
-                connectSingleShot(testSuite->room(), &Room::addedMessages, this,
-                                  &TestManager::doTests);
+                testSuite->room()->getPreviousContent().then(this, &TestManager::doTests);
             }
         });
     });
-    connect(joinJob, &BaseJob::failure, this, [this] {
-        clog << "Failed to join the test room" << endl;
-        finalize();
-    });
 }
 
 void TestManager::onNewRoom(Room* r)
 {
-    clog << "New room: " << r->id().toStdString() << endl
-         << "  Name: " << r->name().toStdString() << endl
-         << "  Canonical alias: " << r->canonicalAlias().toStdString() << endl
-         << endl;
+    clog << "New room: " << r->id().toStdString() << "\n  Name: " << r->name().toStdString()
+         << "\n  Canonical alias: " << r->canonicalAlias().toStdString() << endl;
     connect(r, &Room::aboutToAddNewMessages, r, [r](RoomEventsRange timeline) {
-        clog << timeline.size() << " new event(s) in room "
-             << r->objectName().toStdString() << endl;
+        clog << timeline.size() << " new event(s) in room " << r->objectName().toStdString() << endl;
     });
 }
 
@@ -301,7 +322,7 @@ void TestManager::doTests()
                                   Q_ARG(QByteArray, testName));
     }
     clog << "Tests to do:";
-    for (const auto& test: qAsConst(running))
+    for (const auto& test: std::as_const(running))
         clog << " " << testName(test);
     clog << endl;
     connect(testSuite, &TestSuite::finishedItem, this,
@@ -330,36 +351,33 @@ TEST_IMPL(loadMembers)
     // It's not exactly correct because an arbitrary server might not support
     // lazy loading; but in the absence of capabilities framework we assume
     // it does.
-    if (targetRoom->users().size() >= targetRoom->joinedCount()) {
+    if (targetRoom->joinedMembers().size() >= targetRoom->joinedCount()) {
         clog << "Lazy loading doesn't seem to be enabled" << endl;
         FAIL_TEST();
     }
     targetRoom->setDisplayed();
     connect(targetRoom, &Room::allMembersLoaded, this, [this, thisTest] {
-        FINISH_TEST(targetRoom->users().size() >= targetRoom->joinedCount());
+        FINISH_TEST(targetRoom->joinedMembers().size() >= targetRoom->joinedCount());
     });
     return false;
 }
 
 TEST_IMPL(sendMessage)
 {
-    auto txnId = targetRoom->postPlainText("Hello, "_ls % origin % " is here"_ls);
+    auto txnId = targetRoom->postPlainText("Hello, "_L1 % origin % " is here"_L1);
     if (!validatePendingEvent<RoomMessageEvent>(txnId)) {
         clog << "Invalid pending event right after submitting" << endl;
         FAIL_TEST();
     }
-    connectUntil(targetRoom, &Room::pendingEventAboutToMerge, this,
-        [this, thisTest, txnId](const RoomEvent* evt, int pendingIdx) {
-            const auto& pendingEvents = targetRoom->pendingEvents();
-            Q_ASSERT(pendingIdx >= 0 && pendingIdx < int(pendingEvents.size()));
-
-            if (evt->transactionId() != txnId)
-                return false;
-
-            FINISH_TEST(is<RoomMessageEvent>(*evt) && !evt->id().isEmpty()
-                        && pendingEvents[size_t(pendingIdx)]->transactionId()
-                               == evt->transactionId());
-        });
+    targetRoom->whenMessageMerged(txnId).then(this, [this, thisTest, txnId](const RoomEvent& evt) {
+        const auto pendingIt = targetRoom->findPendingEvent(txnId);
+        if (pendingIt == targetRoom->pendingEvents().end()) {
+            clog << "Pending event not found at the moment of local echo merging\n";
+            FAIL_TEST();
+        }
+        FINISH_TEST(evt.is<RoomMessageEvent>() && !evt.id().isEmpty()
+                    && txnId == (*pendingIt)->transactionId() && txnId == evt.transactionId());
+    });
     return false;
 }
 
@@ -378,7 +396,7 @@ TEST_IMPL(sendReaction)
         FAIL_TEST();
     }
 
-    const auto key = QStringLiteral("+1");
+    const auto key = u"+1"_s;
     const auto txnId = targetRoom->postReaction(targetEvtId, key);
     if (!validatePendingEvent<ReactionEvent>(txnId)) {
         clog << "Invalid pending event right after submitting" << endl;
@@ -414,13 +432,13 @@ TEST_IMPL(sendFile)
     }
     tf->write("Test");
     tf->close();
-    QFileInfo tfi { *tf };
+    const QFileInfo tfi { *tf };
     // QFileInfo::fileName brings only the file name; QFile::fileName brings
     // the full path
     const auto tfName = tfi.fileName();
     clog << "Sending file " << tfName.toStdString() << endl;
     const auto txnId = targetRoom->postFile(
-        "Test file"_ls, new EventContent::FileContent(tfi));
+        "Test file"_L1, std::make_unique<EventContent::FileContent>(tfi));
     if (!validatePendingEvent<RoomMessageEvent>(txnId)) {
         clog << "Invalid pending event right after submitting" << endl;
         tf->deleteLater();
@@ -445,44 +463,39 @@ TEST_IMPL(sendFile)
             if (id != txnId)
                 return false;
 
-            targetRoom->postPlainText(origin % ": File upload failed: "_ls % error);
+            targetRoom->postPlainText(origin % ": File upload failed: "_L1 % error);
             tf->deleteLater();
             FAIL_TEST();
         });
     return false;
 }
 
-// Can be replaced with a lambda once QtConcurrent is able to resolve return
-// types from lambda invocations (Qt 6 can, not sure about earlier)
-struct DownloadRunner {
-    QUrl url;
-
-    using result_type = QNetworkReply::NetworkError;
-
-    QNetworkReply::NetworkError operator()(int) const
-    {
-        QEventLoop el;
-        QScopedPointer<QNetworkReply, QScopedPointerDeleteLater> reply {
-            NetworkAccessManager::instance()->get(QNetworkRequest(url))
-        };
-        QObject::connect(
-            reply.data(), &QNetworkReply::finished, &el, [&el] { el.exit(); },
-            Qt::QueuedConnection);
-        el.exec();
-        return reply->error();
-    }
-};
+void getResource(const QUrl& url, QScopedPointer<QNetworkReply, QScopedPointerDeleteLater>& r,
+                 QEventLoop& el)
+{
+    r.reset(NetworkAccessManager::instance()->get(QNetworkRequest(url)));
+    QObject::connect(
+        r.data(), &QNetworkReply::finished, &el,
+        [url, &r, &el] {
+            if (r->error() != QNetworkReply::NoError)
+                getResource(url, r, el);
+            else
+                el.exit();
+        },
+        Qt::QueuedConnection);
+}
 
 bool testDownload(const QUrl& url)
 {
-    // Move out actual test from the multithreaded code
-    // to help debugging
-    auto results = QtConcurrent::blockingMapped(QVector<int> { 1, 2, 3 },
-                                                DownloadRunner { url });
-    return std::all_of(results.cbegin(), results.cend(),
-                        [](QNetworkReply::NetworkError ne) {
-                            return ne == QNetworkReply::NoError;
-                        });
+    // The actual test is separate from the download invocation to help debugging
+    const auto results = QtConcurrent::blockingMapped(QVector<int>{ 1, 2, 3 }, [url](int) {
+        thread_local QEventLoop el;
+        thread_local QScopedPointer<QNetworkReply, QScopedPointerDeleteLater> reply{};
+        getResource(url, reply, el);
+        el.exec();
+        return reply->error();
+    });
+    return results == QVector<QNetworkReply::NetworkError>(3, QNetworkReply::NoError);
 }
 
 bool TestSuite::checkFileSendingOutcome(const TestToken& thisTest,
@@ -501,28 +514,18 @@ bool TestSuite::checkFileSendingOutcome(const TestToken& thisTest,
         FAIL_TEST();
     }
 
-    connectUntil(targetRoom, &Room::pendingEventAboutToMerge, this,
-        [this, thisTest, txnId, fileName](const RoomEvent* evt, int pendingIdx) {
-            const auto& pendingEvents = targetRoom->pendingEvents();
-            Q_ASSERT(pendingIdx >= 0 && pendingIdx < int(pendingEvents.size()));
-
-            if (evt->transactionId() != txnId)
-                return false;
-
-            clog << "File event " << txnId.toStdString()
-                 << " arrived in the timeline" << endl;
-            return switchOnType(
-                *evt,
+    targetRoom->whenMessageMerged(txnId).then(
+        this, [this, thisTest, txnId, fileName](const RoomEvent& evt) {
+            clog << "File event " << txnId.toStdString() << " arrived in the timeline" << endl;
+            using EventContent::FileContent;
+            evt.switchOnType(
                 [&](const RoomMessageEvent& e) {
                     // TODO: check #366 once #368 is implemented
-                    FINISH_TEST(
-                        !e.id().isEmpty()
-                        && pendingEvents[size_t(pendingIdx)]->transactionId()
-                               == txnId
-                        && e.hasFileContent()
-                        && e.content()->fileInfo()->originalName == fileName
-                        && testDownload(targetRoom->connection()->makeMediaUrl(
-                            e.content()->fileInfo()->url())));
+                    FINISH_TEST(!e.id().isEmpty() && evt.transactionId() == txnId
+                                && e.has<FileContent>()
+                                && e.get<FileContent>()->originalName == fileName
+                                && testDownload(targetRoom->connection()->makeMediaUrl(
+                                    e.get<FileContent>()->url())));
                 },
                 [this, thisTest](const RoomEvent&) { FAIL_TEST(); });
         });
@@ -534,25 +537,19 @@ DEFINE_SIMPLE_EVENT(CustomEvent, RoomEvent, "quotest.custom", int, testValue,
 
 TEST_IMPL(sendCustomEvent)
 {
-    auto txnId = targetRoom->postEvent(new CustomEvent(42));
-    if (!validatePendingEvent<CustomEvent>(txnId)) {
+    const auto& pendingEventItem = targetRoom->post<CustomEvent>(42);
+    if (!validatePendingEvent<CustomEvent>(pendingEventItem->transactionId())) {
         clog << "Invalid pending event right after submitting" << endl;
         FAIL_TEST();
     }
-    connectUntil(
-        targetRoom, &Room::pendingEventAboutToMerge, this,
-        [this, thisTest, txnId](const RoomEvent* evt, int pendingIdx) {
-            const auto& pendingEvents = targetRoom->pendingEvents();
-            Q_ASSERT(pendingIdx >= 0 && pendingIdx < int(pendingEvents.size()));
-
-            if (evt->transactionId() != txnId)
-                return false;
-
-            return switchOnType(*evt,
-                [this, thisTest, &evt](const CustomEvent& e) {
-                    FINISH_TEST(!evt->id().isEmpty() && e.testValue() == 42);
+    pendingEventItem.whenMerged().then(
+        this, [this, thisTest, txnId = pendingEventItem->transactionId()](const RoomEvent& evt) {
+            evt.switchOnType(
+                [this, thisTest, txnId, &evt](const CustomEvent& e) {
+                    FINISH_TEST(!evt.id().isEmpty() && evt.transactionId() == txnId
+                                && e.testValue() == 42);
                 },
-                [this, thisTest] (const RoomEvent&) { FAIL_TEST(); });
+                [this, thisTest](const RoomEvent&) { FAIL_TEST(); });
         });
     return false;
 
@@ -560,8 +557,7 @@ TEST_IMPL(sendCustomEvent)
 
 TEST_IMPL(setTopic)
 {
-    const auto newTopic = connection()->generateTxnId(); // Just a way to make
-                                                         // a unique id
+    const auto newTopic = connection()->generateTxnId(); // Just a way to make a unique id
     targetRoom->setTopic(newTopic);
     connectUntil(targetRoom, &Room::topicChanged, this,
         [this, thisTest, newTopic] {
@@ -614,7 +610,7 @@ TEST_IMPL(changeName)
     // completes (with server roundtrips etc.) the first rename before
     // redactEvent() even starts, redactEvent() will capture the rename event
     // instead of the join, and likely break changeName() as a result.
-    connectSingleShot(targetRoom, &Room::allMembersLoaded, this, [this, thisTest] {
+    QtFuture::connect(targetRoom, &Room::allMembersLoaded).then([this, thisTest] {
         auto* const localUser = connection()->user();
         const auto& newName = connection()->generateTxnId(); // See setTopic()
         clog << "Renaming the user to " << newName.toStdString()
@@ -622,7 +618,7 @@ TEST_IMPL(changeName)
         localUser->rename(newName, targetRoom);
         connectUntil(
             targetRoom, &Room::aboutToAddNewMessages, this,
-            [this, thisTest, localUser, newName](const RoomEventsRange& evts) {
+            [this, thisTest, localUser, newName](RoomEventsRange evts) {
                 for (const auto& e : evts) {
                     if (const auto* rme = eventCast<const RoomMemberEvent>(e)) {
                         if (rme->stateKey() != localUser->id()
@@ -645,7 +641,7 @@ TEST_IMPL(changeName)
                         localUser->rename(newN);
                         connectUntil(localUser, &User::defaultNameChanged, this,
                                      [this, thisTest, localUser, newN] {
-                                         targetRoom->localUser()->rename({});
+                                         localUser->rename({});
                                          FINISH_TEST(localUser->name() == newN);
                                      });
                         return true;
@@ -660,12 +656,12 @@ TEST_IMPL(changeName)
 TEST_IMPL(showLocalUsername)
 {
     auto* const localUser = connection()->user();
-    FINISH_TEST(!localUser->name().contains("@"_ls));
+    FINISH_TEST(!localUser->name().contains("@"_L1));
 }
 
 TEST_IMPL(addAndRemoveTag)
 {
-    static const auto TestTag = QStringLiteral("im.quotient.test");
+    static const auto TestTag = u"im.quotient.test"_s;
     // Pre-requisite
     if (targetRoom->tags().contains(TestTag))
         targetRoom->removeTag(TestTag);
@@ -675,37 +671,42 @@ TEST_IMPL(addAndRemoveTag)
     // that the signal is emitted, not only that tags have changed; but there's
     // (currently) no way to check that the server has been correctly notified
     // of the tag change.
-    QSignalSpy spy(targetRoom, &Room::tagsChanged);
+    const QSignalSpy spy(targetRoom, &Room::tagsChanged);
     targetRoom->addTag(TestTag);
-    if (spy.count() != 1 || !targetRoom->tags().contains(TestTag)) {
+    if (spy.size() != 1 || !targetRoom->tags().contains(TestTag)) {
         clog << "Tag adding failed" << endl;
         FAIL_TEST();
     }
+    const auto& tagsToRooms = connection()->tagsToRooms();
+    if (!tagsToRooms.contains(TestTag) || !tagsToRooms[TestTag].contains(targetRoom)) {
+        clog << "Tag adding succeeded but the connection doesn't know about it\n";
+        FAIL_TEST();
+    }
     clog << "Test tag set, removing it now" << endl;
     targetRoom->removeTag(TestTag);
-    FINISH_TEST(spy.count() == 2 && !targetRoom->tags().contains(TestTag));
+    FINISH_TEST(spy.size() == 2 && !targetRoom->tags().contains(TestTag));
 }
 
 bool TestSuite::checkDirectChat() const
 {
-    return targetRoom->directChatUsers().contains(connection()->user());
+    return targetRoom->directChatMembers().contains(targetRoom->member(connection()->user()->id()));
 }
 
 TEST_IMPL(markDirectChat)
 {
     if (checkDirectChat())
         connection()->removeFromDirectChats(targetRoom->id(),
-                                            connection()->user());
+                                            connection()->user()->id());
 
-    int id = qRegisterMetaType<DirectChatsMap>(); // For QSignalSpy
+    const auto id = qRegisterMetaType<DirectChatsMap>(); // For QSignalSpy
     Q_ASSERT(id != -1);
 
     // Same as with tags (and unusual for the rest of Quotient), direct chat
     // operations are synchronous.
-    QSignalSpy spy(connection(), &Connection::directChatsListChanged);
+    const QSignalSpy spy(connection(), &Connection::directChatsListChanged);
     clog << "Marking the room as a direct chat" << endl;
-    connection()->addToDirectChats(targetRoom, connection()->user());
-    if (spy.count() != 1 || !checkDirectChat())
+    connection()->addToDirectChats(targetRoom, connection()->user()->id());
+    if (spy.size() != 1 || !checkDirectChat())
         FAIL_TEST();
 
     // Check that the first argument (added DCs) actually contains the room
@@ -717,8 +718,8 @@ TEST_IMPL(markDirectChat)
     }
 
     clog << "Unmarking the direct chat" << endl;
-    connection()->removeFromDirectChats(targetRoom->id(), connection()->user());
-    if (spy.count() != 2 && checkDirectChat())
+    connection()->removeFromDirectChats(targetRoom->id(), connection()->user()->id());
+    if (spy.size() != 2 && checkDirectChat())
         FAIL_TEST();
 
     // Check that the second argument (removed DCs) actually contains the room
@@ -738,18 +739,17 @@ TEST_IMPL(visitResources)
     static UriDispatcher ud;
 
     // This lambda returns true in case of error, false if it's fine so far
-    auto testResourceResolver = [this, thisTest](const QStringList& uris,
-                                                 auto signal, auto* target,
-                                                 QVariantList otherArgs = {}) {
-        int r = qRegisterMetaType<decltype(target)>();
+    const auto testResourceResolver = [this, thisTest](const QStringList& uris, auto signal,
+                                                       auto* target, QVariantList otherArgs = {}) {
+        const auto r = qRegisterMetaType<decltype(target)>();
         Q_ASSERT(r != 0);
         QSignalSpy spy(&ud, signal);
         for (const auto& uriString: uris) {
-            Uri uri { uriString };
+            const Uri uri { uriString };
             clog << "Checking " << uriString.toStdString()
                  << " -> " << uri.toDisplayString().toStdString() << endl;
             if (auto matrixToUrl = uri.toUrl(Uri::MatrixToUri).toDisplayString();
-                !matrixToUrl.startsWith("https://matrix.to/#/"_ls)) {
+                !matrixToUrl.startsWith("https://matrix.to/#/"_L1)) {
                 clog << "Incorrect matrix.to representation:"
                      << matrixToUrl.toStdString() << endl;
             }
@@ -762,13 +762,13 @@ TEST_IMPL(visitResources)
                 FAIL_TEST();
             }
             ud.visitResource(connection(), uriString);
-            if (spy.count() != 1) {
-                clog << "Wrong number of signal emissions (" << spy.count()
+            if (spy.size() != 1) {
+                clog << "Wrong number of signal emissions (" << spy.size()
                      << ')' << endl;
                 FAIL_TEST();
             }
             const auto& emission = spy.front();
-            Q_ASSERT(emission.count() >= 2);
+            Q_ASSERT(emission.size() >= 2);
             if (emission.front().value<decltype(target)>() != target) {
                 clog << "Signal emitted with an incorrect target" << endl;
                 FAIL_TEST();
@@ -795,14 +795,13 @@ TEST_IMPL(visitResources)
             clog << "Empty Matrix URI test failed" << endl;
             FAIL_TEST();
         }
-    if (Uri { QStringLiteral("#") }.isValid()) {
+    if (Uri { u"#"_s }.isValid()) {
         clog << "Bare sigil URI test failed" << endl;
         FAIL_TEST();
     }
-    QUrl invalidUrl { "https://"_ls };
-    invalidUrl.setAuthority("---:@@@"_ls);
-    const Uri matrixUriFromInvalidUrl { invalidUrl },
-        invalidMatrixUri { QStringLiteral("matrix:&invalid@") };
+    QUrl invalidUrl { "https://"_L1 };
+    invalidUrl.setAuthority("---:@@@"_L1);
+    const Uri matrixUriFromInvalidUrl{ invalidUrl }, invalidMatrixUri{ u"matrix:&invalid@"_s };
     if (matrixUriFromInvalidUrl.isEmpty() || matrixUriFromInvalidUrl.isValid()) {
         clog << "Invalid Matrix URI test failed" << endl;
         FAIL_TEST();
@@ -823,57 +822,50 @@ TEST_IMPL(visitResources)
     Q_ASSERT(!eventId.isEmpty());
 
     const QStringList roomUris {
-        roomId, "matrix:roomid/"_ls + roomId.mid(1),
-        "https://matrix.to/#/%21"_ls/*`!`*/ + roomId.mid(1),
-        roomAlias, "matrix:room/"_ls + roomAlias.mid(1),
-        "matrix:r/"_ls + roomAlias.mid(1),
-        "https://matrix.to/#/"_ls + roomAlias,
+        roomId, "matrix:roomid/"_L1 + roomId.mid(1),
+        "https://matrix.to/#/%21"_L1/*`!`*/ + roomId.mid(1),
+        roomAlias, "matrix:room/"_L1 + roomAlias.mid(1),
+        "matrix:r/"_L1 + roomAlias.mid(1),
+        "https://matrix.to/#/"_L1 + roomAlias,
     };
-    const QStringList userUris { userId, "matrix:user/"_ls + userId.mid(1),
-                                 "matrix:u/"_ls + userId.mid(1),
-                                 "https://matrix.to/#/"_ls + userId };
+    const QStringList userUris { userId, "matrix:user/"_L1 + userId.mid(1),
+                                 "matrix:u/"_L1 + userId.mid(1),
+                                 "https://matrix.to/#/"_L1 + userId };
     const QStringList eventUris {
-        "matrix:room/"_ls + roomAlias.mid(1) + "/event/"_ls + eventId.mid(1),
-        "matrix:r/"_ls + roomAlias.mid(1) + "/e/"_ls + eventId.mid(1),
-        "https://matrix.to/#/"_ls + roomId + u'/' + eventId
+        "matrix:room/"_L1 + roomAlias.mid(1) + "/event/"_L1 + eventId.mid(1),
+        "matrix:r/"_L1 + roomAlias.mid(1) + "/e/"_L1 + eventId.mid(1),
+        "https://matrix.to/#/"_L1 + roomId + u'/' + eventId
     };
     // Check that reserved characters are correctly processed.
-    static const auto& joinRoomAlias =
-        QStringLiteral("##/?.@\"unjoined:example.org");
+    static const auto joinRoomAlias = u"##/?.@\"unjoined:example.org"_s;
     static const auto& encodedRoomAliasNoSigil =
-        QString::fromLatin1(QUrl::toPercentEncoding(joinRoomAlias.mid(1), QByteArrayLiteral(":")));
-    static const QString joinQuery { "?action=join"_ls };
+        QString::fromLatin1(QUrl::toPercentEncoding(joinRoomAlias.mid(1), ":"_ba));
+    static const auto joinQuery = u"?action=join"_s;
     // These URIs are not supposed to be actually joined (and even exist,
     // as yet) - only to be syntactically correct
-    static const QStringList joinByAliasUris {
+    static const QStringList joinByAliasUris{
         Uri(joinRoomAlias.toUtf8(), {}, joinQuery.mid(1)).toDisplayString(),
-        "matrix:room/"_ls + encodedRoomAliasNoSigil + joinQuery,
-        "matrix:r/"_ls + encodedRoomAliasNoSigil + joinQuery,
-        "https://matrix.to/#/%23"_ls/*`#`*/ + encodedRoomAliasNoSigil + joinQuery,
-        "https://matrix.to/#/%23"_ls + joinRoomAlias.mid(1) /* unencoded */ + joinQuery
-    };
-    static const auto& joinRoomId = QStringLiteral("!anyid:example.org");
-    static const QStringList viaServers { "matrix.org"_ls, "example.org"_ls };
-    static const auto viaQuery =
-        std::accumulate(viaServers.cbegin(), viaServers.cend(), joinQuery,
-                        [](const QString& q, const QString& s) {
-                            return q + "&via="_ls + s;
-                        });
-    static const QStringList joinByIdUris {
-        "matrix:roomid/"_ls + joinRoomId.mid(1) + viaQuery,
-        "https://matrix.to/#/"_ls + joinRoomId + viaQuery
+        "matrix:room/"_L1 % encodedRoomAliasNoSigil % joinQuery,
+        "matrix:r/"_L1 % encodedRoomAliasNoSigil % joinQuery,
+        "https://matrix.to/#/%23"_L1 /*`#`*/ % encodedRoomAliasNoSigil % joinQuery,
+        "https://matrix.to/#/%23"_L1 % joinRoomAlias.mid(1) /* unencoded */ % joinQuery
     };
+    static const auto joinRoomId = u"!anyid:example.org"_s;
+    static constexpr auto viaServers = std::to_array({ "matrix.org"_L1, "example.org"_L1 });
+    static const auto viaQuery = std::apply(
+        [](const auto&... servers) { return QString((joinQuery % ... % (u"&via="_s % servers))); },
+        viaServers);
+    static const QStringList joinByIdUris{ "matrix:roomid/"_L1 % joinRoomId.mid(1) % viaQuery,
+                                           "https://matrix.to/#/"_L1 % joinRoomId % viaQuery };
     // If any test breaks, the breaking call will return true, and further
     // execution will be cut by ||'s short-circuiting
     if (testResourceResolver(roomUris, &UriDispatcher::roomAction, room())
-        || testResourceResolver(userUris, &UriDispatcher::userAction,
-                                connection()->user())
-        || testResourceResolver(eventUris, &UriDispatcher::roomAction,
-                                room(), { eventId })
-        || testResourceResolver(joinByAliasUris, &UriDispatcher::joinAction,
-                                connection(), { joinRoomAlias })
-        || testResourceResolver(joinByIdUris, &UriDispatcher::joinAction,
-                                connection(), { joinRoomId, viaServers }))
+        || testResourceResolver(userUris, &UriDispatcher::userAction, connection()->user())
+        || testResourceResolver(eventUris, &UriDispatcher::roomAction, room(), { eventId })
+        || testResourceResolver(joinByAliasUris, &UriDispatcher::joinAction, connection(),
+                                { joinRoomAlias })
+        || testResourceResolver(joinByIdUris, &UriDispatcher::joinAction, connection(),
+                                { joinRoomId, QStringList(viaServers.cbegin(), viaServers.cend()) }))
         return true;
     // TODO: negative cases
     FINISH_TEST(true);
@@ -900,86 +892,68 @@ void TestManager::conclude()
     // Clean up the room (best effort)
     auto* room = testSuite->room();
     room->setTopic({});
-    room->localUser()->rename({});
+    c->user()->rename({});
 
-    QString succeededRec { QString::number(succeeded.size()) % " of "_ls
-                           % QString::number(succeeded.size() + failed.size()
-                                             + running.size())
-                           % " tests succeeded"_ls };
-    QString plainReport = origin % ": Testing complete, "_ls % succeededRec;
-    QString color = failed.empty() && running.empty() ? "00AA00"_ls : "AA0000"_ls;
-    QString htmlReport = origin % ": <strong><font data-mx-color='#"_ls % color
-                         % "' color='#"_ls % color
-                         % "'>Testing complete</font></strong>, "_ls % succeededRec;
+    const QString succeededRec{ QString::number(succeeded.size()) % " of "_L1
+                                % QString::number(succeeded.size() + failed.size() + running.size())
+                                % " tests succeeded"_L1 };
+    QString plainReport = origin % ": Testing complete, "_L1 % succeededRec;
+    const QString color = failed.empty() && running.empty() ? "00AA00"_L1 : "AA0000"_L1;
+    QString htmlReport = origin % ": <strong><font data-mx-color='#"_L1 % color
+                         % "' color='#"_L1 % color
+                         % "'>Testing complete</font></strong>, "_L1 % succeededRec;
     if (!failed.empty()) {
         QByteArray failedList;
-        for (const auto& f : qAsConst(failed))
+        for (const auto& f : std::as_const(failed))
             failedList += ' ' + f;
-        plainReport += "\nFAILED:"_ls + QString::fromUtf8(failedList);
-        htmlReport += "<br><strong>Failed:</strong>"_ls + QString::fromUtf8(failedList);
+        plainReport += "\nFAILED:"_L1 + QString::fromUtf8(failedList);
+        htmlReport += "<br><strong>Failed:</strong>"_L1 + QString::fromUtf8(failedList);
     }
     if (!running.empty()) {
         QByteArray dnfList;
-        for (const auto& r : qAsConst(running))
+        for (const auto& r : std::as_const(running))
             dnfList += ' ' + r;
-        plainReport += "\nDID NOT FINISH:"_ls + QString::fromUtf8(dnfList);
-        htmlReport += "<br><strong>Did not finish:</strong>"_ls + QString::fromUtf8(dnfList);
+        plainReport += "\nDID NOT FINISH:"_L1 + QString::fromUtf8(dnfList);
+        htmlReport += "<br><strong>Did not finish:</strong>"_L1 + QString::fromUtf8(dnfList);
     }
 
     auto txnId = room->postHtmlText(plainReport, htmlReport);
     // Now just wait until all the pending events reach the server
-    connectUntil(room, &Room::messageSent, this,
-        [this, txnId, room, plainReport] (const QString& sentTxnId) {
-            if (sentTxnId != txnId)
-                return false;
-            const auto& pendingEvents = room->pendingEvents();
-            if (auto c = std::count_if(pendingEvents.cbegin(),
-                                       pendingEvents.cend(),
-                                       [](const PendingEventItem& pe) {
-                                           return pe.deliveryStatus()
-                                                  < EventStatus::ReachedServer;
-                                       });
-                c > 0) {
-                clog << "Events to reach the server: " << c
-                     << ", not leaving yet" << endl;
-                return false;
-            }
+    connectUntil(room, &Room::messageSent, this, [this, txnId, room, plainReport] {
+        const auto& pendingEvents = room->pendingEvents();
+        if (const auto stillFlyingCount =
+                std::ranges::count_if(pendingEvents,
+                                      [](const PendingEventItem& pe) {
+                                          return pe.deliveryStatus() < EventStatus::ReachedServer;
+                                      });
+            stillFlyingCount > 0) {
+            clog << "Events to reach the server: " << stillFlyingCount << ", not leaving yet\n";
+            return false;
+        }
 
-            clog << "Leaving the room" << endl;
-            // TODO: Waiting for proper futures to come so that it could be:
-//           room->leaveRoom()
-//           .then(this, &TestManager::finalize); // Qt-style or
-//           .then([this] { finalize(); }); // STL-style
-            auto* job = room->leaveRoom();
-            connect(job, &BaseJob::result, this, [this, job,plainReport] {
-                Q_ASSERT(job->status().good());
-                finalize();
-                // Still flying, as the exit() connection in finalize() is queued
-                clog << plainReport.toStdString() << endl;
-            });
-            return true;
-        });
+        clog << "Leaving the room" << endl;
+        room->leaveRoom().then(this, std::bind_front(&TestManager::finalize, this, plainReport));
+        return true;
+    });
 }
 
-void TestManager::finalize()
+void TestManager::finalize(const QString& lastWords)
 {
-    if (!c->isUsable() || !c->isLoggedIn()) {
+    if (!c->isLoggedIn()) {
         clog << "No usable connection reached" << endl;
         QCoreApplication::exit(-2);
         return; // NB: QCoreApplication::exit() does return to the caller
     }
     clog << "Logging out" << endl;
-    c->logout();
-    connect(
-        c, &Connection::loggedOut, this,
-        [this] {
+    c->logout().then(
+        this, [this, lastWords] {
+            clog << lastWords.toStdString() << endl;
             QCoreApplication::exit(!testSuite ? -3
                                    : succeeded.empty() && failed.empty()
                                            && running.empty()
                                        ? -4
-                                       : failed.size() + running.size());
-        },
-        Qt::QueuedConnection);
+                                       : static_cast<int>(failed.size() + running.size()));
+        });
 }
 
 int main(int argc, char* argv[])
